/** 
 * @file test/hashtest.c
 * @brief testcase for util/semaphore.c
 **/

#include "gnunet_util.h"

static Mutex lock;

static Semaphore * sem;

static volatile int sv;

static volatile int tv;

static IPC_Semaphore * ipc;

static void lockIt() {
  sv = 0;
  fprintf(stderr, ".");
  while (sv == 0) ;
  MUTEX_LOCK(&lock);
  sv = 1;
  MUTEX_UNLOCK(&lock);
  sv = 2;
  tv = 2;
}

static void bigStack() {
  int i;
  char big[1024 * 100];

  fprintf(stderr, ".");
  for (i=0;i<1024*100;i++)
    big[i] = (char) i;
}

static int testPTHREAD_CREATE() {
  pthread_t pt;

  sv = -1;
  fprintf(stderr, ".");
  MUTEX_CREATE(&lock);
  PTHREAD_CREATE(&pt,
		 (PThreadMain)&lockIt,
		 NULL,
		 1024);
  while (tv != 2) sv = 1;
  MUTEX_DESTROY(&lock);
  PTHREAD_CREATE(&pt,
		 (PThreadMain)&bigStack,
		 NULL,
		 1024*100 + 25000); /* fails by segfault */
  return 0;
}

static int testMutex() {  
  pthread_t pt;
  MUTEX_CREATE(&lock);

  sv = 1;
  tv = 0;
  PTHREAD_CREATE(&pt,
		 (PThreadMain)&lockIt,
		 NULL,
		 1024);
  while (sv == 1) ;
  MUTEX_LOCK(&lock);
  sv = 5; /* release lockIt from while sv==0 loop,
	     blocks it on lock */
  fprintf(stderr, ".");
  sleep(1);
  if (sv != 5) {
    MUTEX_UNLOCK(&lock);
    while (tv != 2) ;
    MUTEX_DESTROY(&lock);
    printf("MUTEX test failed at %s:%u\n",
	   __FILE__, __LINE__);
    return 1; /* error */
  } else {
    MUTEX_UNLOCK(&lock);
    while (tv != 2) ;
    MUTEX_DESTROY(&lock);
    return 0; /* ok */
  } 
}

static int testRecursiveMutex() {
  int i;

  fprintf(stderr, ".");
  MUTEX_CREATE_RECURSIVE(&lock);
  for (i=0;i<50;i++)
    MUTEX_LOCK(&lock);
  for (i=0;i<50;i++)
    MUTEX_UNLOCK(&lock);
  MUTEX_DESTROY(&lock);
  return 0; /* ok -- fails by hanging!*/
} 

static void semUpDown() {
  int i;

  fprintf(stderr, ".");
  for (i=0;i<42;i++)
    SEMAPHORE_DOWN(sem); /* fails by blocking */
  if (SEMAPHORE_DOWN_NONBLOCKING(sem) != SYSERR) {
    SEMAPHORE_FREE(sem);
    printf("SEMAPHORE_DOWN_NONBLOCKING failed at %s:%u\n"
	   "Testcase deadlocked.\n",
	   __FILE__, __LINE__);
    return; /* will halt testcase! */
  }
  for (i=0;i<42;i++)
    SEMAPHORE_UP(sem);
}

static int testSemaphore() {
  int i;
  pthread_t pt;

  sem = SEMAPHORE_NEW(42);
  fprintf(stderr, ".");
  for (i=0;i<42;i++)
    SEMAPHORE_DOWN(sem); /* fails by blocking */
  if (SEMAPHORE_DOWN_NONBLOCKING(sem) != SYSERR) {
    SEMAPHORE_FREE(sem);
    printf("SEMAPHORE_DOWN_NONBLOCKING failed at %s:%u\n",
	   __FILE__, __LINE__);
    return 1;
  }
  for (i=0;i<42;i++)
    SEMAPHORE_UP(sem);
  for (i=0;i<42;i++)
    if (OK != SEMAPHORE_DOWN_NONBLOCKING(sem)) {
      SEMAPHORE_FREE(sem);
      printf("SEMAPHORE_DOWN_NONBLOCKING failed at %s:%u\n",
	     __FILE__, __LINE__);
      return 1;
    }
  if (SEMAPHORE_DOWN_NONBLOCKING(sem) != SYSERR) {
    SEMAPHORE_FREE(sem);
    printf("SEMAPHORE_DOWN_NONBLOCKING failed at %s:%u\n",
	   __FILE__, __LINE__);
    return 1;
  }
  fprintf(stderr, ".");
  PTHREAD_CREATE(&pt,
		 (PThreadMain)&semUpDown,
		 NULL,
		 1024);
  for (i=0;i<42;i++)
    SEMAPHORE_UP(sem);
  for (i=0;i<42;i++)
    SEMAPHORE_DOWN(sem);
  if (SEMAPHORE_DOWN_NONBLOCKING(sem) != SYSERR) {
    SEMAPHORE_FREE(sem);
    printf("SEMAPHORE_DOWN_NONBLOCKING failed at %s:%u\n",
	   __FILE__, __LINE__);
    return 1;
  }
  return 0;
}

static int testIPCSemaphore() {
  int me;
  int cnt;
  int i;
  int j;
  FILE * fd;  

  me = fork();

  remove("/tmp/gnunet_ipc_xchange");
  remove("/tmp/gnunet_ipc_semtest");
  ipc = IPC_SEMAPHORE_NEW("/tmp/gnunet_ipc_semtest",
			  0);
  for (cnt=0;cnt<3;cnt++) {
    if (me == 0) {
      for (i=0;i<6;i++) {
	IPC_SEMAPHORE_DOWN(ipc);
	fd = fopen("/tmp/gnunet_ipc_xchange",
		   "r+");
	if (fd == NULL) {
	  printf("Could not open testfile for reading: %s\n",
		 strerror(errno));
	  return 1;
	}
	fseek(fd, 4*i, SEEK_SET);
	me = fread(&j, 4, 1, fd);
	if (me != 1) {
	  printf("Could not read from testfile: %d - %s\n",
		 me,
		 strerror(errno));
	  return 1;
	}
	fclose(fd);  
	if (j != i+cnt) {
	  IPC_SEMAPHORE_FREE(ipc);
	  printf("IPC test failed at cnt=%d i=%d j=%d %s:%u\n",
		 cnt, i, j, __FILE__, __LINE__);
	  return 1;
	} else
	  fprintf(stderr, ".");
      } 
      remove("/tmp/gnunet_ipc_xchange");     
      me = 1;
    } else {
      for (i=0;i<6;i++) {
	sleep(1);
	fd = fopen("/tmp/gnunet_ipc_xchange",
		   "w+");
	if (fd == NULL) {
	  printf("Could not open testfile for writing: %s\n",
		 strerror(errno));
	  return 1;
	}
	fseek(fd, 4*i, SEEK_SET);
	j=cnt+i;
	if (1 != fwrite(&j, 4, 1, fd)) {
	  printf("Could not write to testfile: %s\n",
		 strerror(errno));
	  return 1;
	}
	fclose(fd);
	IPC_SEMAPHORE_UP(ipc);
      }      
      sleep(2); /* give reader ample time to finish */
      me = 0;
    }
  }
  IPC_SEMAPHORE_FREE(ipc);
  remove("/tmp/gnunet_ipc_xchange");
  if (me == cnt % 2)
    exit(0);
  return 0;
}


int main(int argc, char * argv[]){
  int ret = 0;

  ret += testPTHREAD_CREATE();
  ret += testMutex();
  ret += testRecursiveMutex();
  ret += testSemaphore();
  ret += testIPCSemaphore();
  fprintf(stderr, "\n");
  return ret;
}
